{$APPTYPE CONSOLE}

program SignExtend;

uses
  Windows,
  Sysutils
  ;


{$ifdef VER90}
type
  ptrint  = longint;
  ptruint = dword;
{$endif}

type
  ptrint  = NativeInt;
  ptruint = NativeUint;

type
  PItem    = ^TItem;
  TItem    = packed record
    ItemData : ptruint;
    ItemText : packed array[0..0] of byte;    // variable size array
  end;

  PIndex   = ^TIndex;
  TIndex   = packed array[1..1] of PItem;     // variable size array

  PColIndex = ^TColIndex;
  TColIndex = packed array[0..0] of TIndex;   // variable size array

var
  MemBlock    : pointer;         // address of memory block (constant)

  LineAreaPtr : PItem;           // starts at memblock, grows upwards
  LineIdxPtr  : PColIndex;       // starts at memblock + PAGE, grows downwards

  RowIdx      : integer;         // row index
  I           : integer;         // column index

  p           : pointer;

const
  PAGE_SIZE   = 4096;

begin
  writeln;

  MemBlock := VirtualAlloc(nil,
                           16 * 1024,
                           MEM_RESERVE,
                           PAGE_READWRITE);

  if MemBlock = nil then
  begin
    writeln('Failed to reserve a 16KB memory block. Program terminating.');
    halt(1);
  end;

  writeln('16 KB reserved at address : ', IntToHex(ptruint(MemBlock), 0));

  // make the first page read/write

  MemBlock := VirtualAlloc(MemBlock,
                           1,                            // 1 byte = entire page
                           MEM_COMMIT,
                           PAGE_READWRITE);

  if MemBlock = nil then
  begin
    writeln('Failed to commit the first page in the reserved block.' +
            'Program terminating');
    halt(2);
  end;

  writeln('4KB have been commited at : ', IntToHex(ptruint(MemBlock), 0));

  I      := 0;
  RowIdx := -3;

  // make the Line index pointer point one byte _after_ the allocated block

  LineIdxPtr  := PColIndex(pansichar(MemBlock) + PAGE_SIZE);

  writeln;
  writeln('Top of buffer (inaccessible) is at address : ',
          IntToHex(ptruint(LineIdxPtr), 0));

  writeln('Accessible byte range is                   : ',
           IntToHex(ptruint(MemBlock), 0),
          ' through ',
           IntToHex(ptruint(pansichar(MemBlock) + PAGE_SIZE - 1), 0));


  // make the LineArea pointer point to the beginning of the block

  LineAreaPtr := PItem(MemBlock);

  writeln;
  writeln('RowIdx : ', RowIdx);
  writeln('I      : ', I);

  // calculate what the value of LineIdxPtr[RowIdx][I + 1] should be

  p := @LineIdxPtr^[RowIdx][I + 1];

  writeln;
  writeln('address LineIdxPtr^[RowIdx][I + 1] is      : ',
          IntToHex(ptruint(p), 0));

  if p < (pansichar(MemBlock) + PAGE_SIZE) then
  begin
    writeln('pointer value is below the top of the buffer : ok');
  end
  else
  begin
    writeln('pointer value is NOT below the top of the buffer.');
  end;

  // try to save the pointer - causes access violation in 64bit

  {$ifdef WIN64}
    //writeln;
    //writeln('next statement will cause an access violation');
    //writeln('press enter/return to continue.');
    //readln;
    //writeln;
  {$endif}

  LineIdxPtr^[RowIdx][I + 1] := LineAreaPtr;

  writeln;
  writeln('press enter/return to end this program.');
  readln;
end.

// (from GDB) - 32bit code generated  (this works)

SignExtend.lpr:132                LineIdxPtr^[RowIdx][I + 1] := LineAreaPtr;
0040198F 8b1520804100             mov    0x418020,%edx
00401995 a130804100               mov    0x418030,%eax
0040199A 8b0d40804100             mov    0x418040,%ecx
004019A0 8d0482                   lea    (%edx,%eax,4),%eax
004019A3 8b1510804100             mov    0x418010,%edx
004019A9 891488                   mov    %edx,(%eax,%ecx,4)


// (from GDB) - 64bit code generated (this does NOT work)

SignExtend.lpr:132                        LineIdxPtr^[RowIdx][I + 1] := LineAreaPtr;
0000000100001B02 488b0517e50100           mov    0x1e517(%rip),%rax        # 0x100020020
0000000100001B09 8b1521e50100             mov    0x1e521(%rip),%edx        # 0x100020030
0000000100001B0F 48630d2ae50100           movslq 0x1e52a(%rip),%rcx        # 0x100020040
0000000100001B16 488d04d0                 lea    (%rax,%rdx,8),%rax
0000000100001B1A 488b15efe40100           mov    0x1e4ef(%rip),%rdx        # 0x100020010
0000000100001B21 488914c8                 mov    %rdx,(%rax,%rcx,8)

// (from VS2017) - 64bit code

  LineIdxPtr^[RowIdx][I + 1] := LineAreaPtr;
0000000100001B02 48 8B 05 17 E5 01 00 mov         rax,qword ptr [main+1EA70h (0100020020h)]
0000000100001B09 8B 15 21 E5 01 00    mov         edx,dword ptr [main+1EA80h (0100020030h)]
0000000100001B0F 48 63 0D 2A E5 01 00 movsxd      rcx,dword ptr [main+1EA90h (0100020040h)]
0000000100001B16 48 8D 04 D0          lea         rax,[rax+rdx*8]
0000000100001B1A 48 8B 15 EF E4 01 00 mov         rdx,qword ptr [main+1EA60h (0100020010h)]
0000000100001B21 48 89 14 C8          mov         qword ptr [rax+rcx*8],rdx


the problem is at

0000000100001B09 8b1521e50100             mov    0x1e521(%rip),%edx        # 0x100020030

where RowIdx (declared as in integer) is moved into edx (instead of being moved
into rdx with sign extension.)

this causes the instruction

0000000100001B16 488d04d0                 lea    (%rax,%rdx,8),%rax

to compute an incorrect address because the value in rdx is not negative (but it
should be)